/*================================================================*
 * File: prune.c
 * Author: L. Rautiainen
 *
 *   1-2.1998: Initial coding.
 * 15.02.1998: Bugs fixed.
 * 21.02.1998: Simplified. 
 * 11.04.1998: Jorma Sajaniemi: Added version number        
 * 07.05.1998: Bugs fixed.
 * 11.06.1998: Spelling mistakes repaired. (not all) 
 * 17.10.1998: Bugs fixed.
 * 04.12.1998: Little memory allocation fix.
 *
 * Laboratory Project on Data Structures and Algorithms
 *
 * Generally speaking this program produces an HTML (Hyper Text 
 * Mark-up Language) document by reading another HTML document. 
 * To make it little more interesting, principal purpose is to 
 * remove some parts from the original document.
 *
 * Program can get set of character strings, a file name 
 * containing set of character strings and a file name containing 
 * set of unvisible entities as command line parameters. There is 
 * also a name of source and destination file introduced in 
 * command line. In command line and files, character strings can 
 * be in arbitrary format, of course within the limits of used 
 * hardware and operating system. Upper and lower characters are
 * equivalents between oneselves. Unvisible entity means a 
 * character code which does not correspond a visible character.
 * For example &nbsp; is a space bar.
 *
 * Programs purpose is to produce from already existing HTML 
 * document an HTML document about which has removed some parts. 
 * A destination file contains a source file after execution of 
 * program. However so that all information between <FONT> and 
 * </FONT> tags has removed if from definition part in question 
 * can be found some of those character strings fed as input. 
 * Also information between two <FONT>...</FONT> areas is 
 * removed if it does not contain visible text in other words 
 * visible characters or entities. In this case images are
 * considered unvisible.
 *
 *
 * EXIT CODES for this program are as follows.
 * exit(1) - Improper arguments.
 * exit(2) - File read error.
 * exit(5) - Out of memory.
 *===============================================================*/

/* libraries */
#include <stdio.h>

/* constant and datatype definitions  */
#define VERSION       "1.2"      /* version number */ 
#define READONLY      "r"        /* open for read only */
#define WRITEONLY     "w"        /* open for write only */
#define OPEN_TAG      '<'        /* start of HTML tag */
#define CLOSE_TAG     '>'        /* end of HTML tag */
#define END_TAG       '/'        /* HTML closing tag */
#define FONT_TAG      "FONT"     /* font tag */
#define OPEN_ENTITY   '&'        /* start of HTML entity */
#define CLOSE_ENTITY  ';'        /* end of HTML entity */
#define FALSE         0          /* boolean false value */
#define TRUE          1          /* boolean true value */
#define BUFSIZE       1500       /* buffer size */
#define HYPHEN        '-'        /* hyphen */
#define BLOCKSIZ      16         /* mem allocation block size */
#define STREND        '\0'       /* end of string marker */
#define NEWLINE       '\n'       /* new line */
#define SPACE         ' '        /* space bar */

typedef int Boolean;             /* logical values and types */

/* dynamic buffer structure */
typedef struct buffernode {
                char              buffer[BUFSIZE];
                int               index;
                struct buffernode *next;
                } BUFFEROBJ;
typedef BUFFEROBJ *P_BUFFEROBJ; /* pointer to buffernode */

/* dynamic list structure */
typedef struct listobject {
                char              *string;
                struct listobject *next;
                } LISTOBJ;
typedef LISTOBJ *P_LISTOBJ; /* pointer to listobject */

/* dynamic binary tree structure */
typedef struct treenode {
                char            *string;
                struct treenode *left;
                struct treenode *right;
                } TREENODE;
typedef TREENODE *P_TREENODE; /* pointer to treenode */

/* global variables */

/* pointer initializations */
P_LISTOBJ HEAD     = NULL;
P_LISTOBJ TAIL     = NULL;
P_TREENODE ROOT    = NULL;
P_BUFFEROBJ FIRST  = NULL;
P_BUFFEROBJ LAST   = NULL;

char *programname; /* pointer to programname - argv[0] */

/* errors prune contains */
char *error[] = {
        "Insufficient memory.\n",
        "illegal option %s\n",
        "cannot open %s\n" };

/* inner entities */
char *entity[] = {
           "hairsp",
           "emsp14",
           "numsp",
           "emsp13",
           "ensp",
           "nbsp",
           "puncsp",
           "emsp",
           "thinsp",
           "the_end"};

/* function prototypes used in main program */
void usage();
void syserror(int errorcode, char *message, char *argument);
int strcomp(register char *str1, register char *str2);
Boolean tree_search(char *str);
Boolean list_search(char *str);
char *gstrf(FILE *srcfp, char ENDCHAR);
char *gstrfstr(char *str, int from, int to);
void add_list(char *str);
void add_buffer(char character);
void add_tree(char *str);
void add_tag_buffer(char open, char close, char *string, int length);
void buffer_file(FILE *destfp, Boolean b);
void load_entities();
void load_list(FILE *srcfp);
void load_tree(FILE *srcfp);
P_TREENODE freetree(P_TREENODE node);
void freelist(P_LISTOBJ obj);

/* main */

main(int argc, char *argv[]) 
{
   FILE *srcfp;    /* pointer to source file */
   FILE *destfp;   /* pointer to destination file */

   char *option;
   char a_char;
   char *string;
   char *tag;

   unsigned int length;
   unsigned int i;
   unsigned int tag_count = 0; /* font tag counter */

   Boolean visible = TRUE;
   Boolean to_buf  = TRUE;
   Boolean just_removed = FALSE;

   programname = argv[0];

   load_entities(); /* load inner entities. */

   /* next while-statement is repeated as long as there's
      options left and first character is hyphen.. */

   while((--argc>0)&&((*++argv)[0] == HYPHEN)){
      if(strlen(*argv) == 2){        
         option = argv[0]+1;         /* next char from hyphen */
         if(*++argv == NULL) usage();
         switch(*option){
            case 's': /* from command line to list */
                      add_list(*argv);
                      break;
            case 'd': /* from file to dynamic list */
                      if((srcfp = fopen(*argv, READONLY)) == NULL)
                         syserror(2,error[2],*argv);
                      else load_list(srcfp);
                      fclose(srcfp);
                      break;
            case 'e': /* from file to binary tree */
                      if((srcfp = fopen(*argv, READONLY)) == NULL)
                         syserror(2,error[2],*argv);
                      else load_tree(srcfp);
                      fclose(srcfp);
                      break;
            default:  syserror(1,error[1],option);
         } /* switch */
      } /* if */
      else syserror(1,error[1],*argv);
   --argc;
   } /* while */

   /* here is only two parameters left (source and 
      destination file names) */

   if(*argv == NULL) usage(); /* source file */
   else if((srcfp = fopen(*argv,READONLY)) == NULL)
           syserror(2,error[2],*argv);
   if(*++argv == NULL) usage(); /* destination file */
   else if((destfp = fopen(*argv,WRITEONLY)) == NULL)
           syserror(2,error[2],*argv);

   /* more interesting part... */

   while((a_char = getc(srcfp)) != EOF){ 
      switch(a_char){                          
         case OPEN_TAG:                        
            string = gstrf(srcfp,CLOSE_TAG);   /* inside tags */
            length = strlen(string);           /* string length */
            switch(string[0]){                 /* first char from string */
               case END_TAG:                   
                  if(to_buf)                   /* add tag to buffer */
                     add_tag_buffer(OPEN_TAG,CLOSE_TAG,string,length);
                  if(!to_buf&&length == 5){
                     tag = gstrfstr(string,1,4);
                     if(!strcomp(tag,FONT_TAG))
                        if(!to_buf&&tag_count == 1){
                           to_buf = TRUE;      /* here when farthest */
                           visible = FALSE;    /* removable /font tag */
                           tag_count = 0;
                           just_removed = TRUE;
                        }
                        else tag_count--;      /* removable font tag */
                     free(tag);
                  }
                  break;
               default:                        
                  if(length>4){                
                     tag = gstrfstr(string,0,3);
                     if(!strcomp(tag,FONT_TAG)){ 
                        if(!to_buf) tag_count++;
                        else{
                        free(tag);
                        tag = gstrfstr(string,5,length);
                        if(list_search(tag)){
                        /* empty buffer to file */
                           buffer_file(destfp,visible);
                           tag_count++;
                           to_buf = FALSE;
                        } /* fourth if */
                        } /* third if's else */
                     } /* second if */
                     free(tag);
                  } /* first if */
                  if(to_buf)
                     add_tag_buffer(OPEN_TAG,CLOSE_TAG,string,length);
            } /* 2nd switch */
            free(string);
            break;
         case OPEN_ENTITY: 
            string = gstrf(srcfp,CLOSE_ENTITY);
            length = strlen(string);
            /* if binary tree does not contain character
               string it's a visible character */
            if(!tree_search(string)) visible = TRUE;
               if(to_buf)
                  add_tag_buffer(OPEN_ENTITY,CLOSE_ENTITY,string,length);
            free(string);
            break;
         default: if(to_buf && (a_char != NEWLINE) && (a_char != SPACE)){
                  /* NEWLINE and SPACE is unvisible */
                     add_buffer(a_char);
                     visible = TRUE;
                  }
                  /* if just removed and next a character is NEWLINE
                     it does not belong to buffer */
                  else if(to_buf && just_removed && a_char == NEWLINE)
                          ;
                  else if(to_buf) add_buffer(a_char);
                  if(just_removed) just_removed = FALSE;
      } /* 1st switch */
   } /* while */
   visible = TRUE; /* making sure the rest of buffer */
   buffer_file(destfp,visible);

   fclose(srcfp);
   fclose(destfp);

   freetree(ROOT);
   freelist(HEAD); 
} /* main */

/* subroutines */

/*===============================================================*
 * Function: usage
 * Parameters: Nothing.
 * Returns: Nothing.
 *
 * Usage of prune, like directions. Just a couple of printouts.
 *===============================================================*/
void usage()
{
   printf("%s (version %s) usage: %s [options] ", programname, 
          VERSION, programname);
   printf("source destination \n");
   printf("\noptions:\n");
   printf("-s string    add character string\n");
   printf("-d filename  load character strings from a file\n");
   printf("-e filename  load entities from a file\n");

   exit(1);
}

/*===============================================================*
 * Function: Viseversa
 * Parameters: A character.
 * Returns: A lower character if upper and vise versa.
 * 
 * Very simple if-then-else structure using a couple of character 
 * string functions.
 *===============================================================*/
char viseversa(char c)
{
   return(c = isupper(c) ? tolower(c) : toupper(c));
}

/*===============================================================*
 * Function: strcomp
 * Parameters: Two character string constant.
 * Returns: An int value that is
 *          <  0 <=> if str1 <  str2
 *          == 0 <=> if str1 == str2
 *          >  0 <=> if str1 >  str2
 *
 * The string comparison starts with the first character in each 
 * string and continues with subsequent characters until the 
 * corresponding characters differ or until the end of the strings 
 * is reached. Upper and lower characters are equivalents.
 *===============================================================*/
int strcomp(register char *str1, register char *str2)
{
   char chr1;
   char chr2;

   while((*str1 == *str2) || (viseversa(*str1) == *str2))
      if(*str1 == STREND) return(0);
      else{*++str1;*++str2;}
   /* avoiding side effects... */
   chr1 = isupper(*str1)?*str1:toupper(*str1);
   chr2 = isupper(*str2)?*str2:toupper(*str2);
   /* if done like *chr1 = isupper(*str...
      the result would not be appropriate... */
   return(chr1-chr2);
}

/*===============================================================*
 * Function: syserror
 * Parameters: An errorcode which is
 *             1: when improper arguments.  
 *             2: when file read error.
 *             5: when out of memory.
 *             a message and an argument which are character string 
 *             constants.
 * Returns: Nothing.
 *           
 * Informs mistakes made by user or problems which program may have. 
 *===============================================================*/
void syserror(int errorcode, char *message, char *argument)
{
   printf("%s[%d]:", programname, errorcode);
   printf(message, argument);

   exit(errorcode);
}

/*===============================================================*
 * Function: new_node
 * Parameters: A character string constant.
 * Returns: Allocated new treenode for binary tree.
 *
 * Memory allocation by using malloc(); function. Nothing serious. 
 * Only thing should be kept in mind is that node->string is also 
 * a pointer so it is sufficient reason to allocate memory also for 
 * it. Function gets a pointer to character string as a parameter. 
 * After memory allocations the string will be copied to 
 * node->string and node's sons are set as nils. Finally allocated 
 * node is returned.
 *===============================================================*/
P_TREENODE new_node(char *str)
{
   char       *malloc();
   P_TREENODE node;

   node = (P_TREENODE) malloc (sizeof(TREENODE));
   if(node == NULL) syserror(5,error[0],"");
   node->string = malloc(strlen(str)+1);
   /* + 1 is needed to accommodate the terminating null character */
   if(node->string == NULL) syserror(5,error[0],"");
   strcpy(node->string,str);
   node->left = node->right = NULL;
   return(node);
}

/*===============================================================*
 * Function: add_tree
 * Parameters: A character string constant.
 * Returns: Nothing. 
 *
 * Purpose is to add string into binary tree. At first current 
 * node is set to point at root. If there is no root let's use a
 * new_node(); function which builds one. The latter part is non-
 * recursive add tree algorithm and it's repeated as long as 
 * addition is executed. Because P_TREENODE is a pointer to global
 * structure, we don't have to return anything.
 *===============================================================*/
void add_tree(char *str)
{
   P_TREENODE new_node();
   P_TREENODE CURRENT;
   Boolean    added = FALSE;

   CURRENT = ROOT;
   if(CURRENT == NULL) ROOT = CURRENT = new_node(str);
   else{
      while(!added){
         if(strcomp(str, CURRENT->string) < 0)
            if(CURRENT->left == NULL){
               CURRENT = CURRENT->left = new_node(str);
               added = TRUE;
            }
            else CURRENT = CURRENT->left;
         else
         if(strcomp(str, CURRENT->string) > 0)
            if(CURRENT->right == NULL){
               CURRENT = CURRENT->right = new_node(str);
               added = TRUE;
            }
            else CURRENT = CURRENT->right;
         else added = TRUE;  /* str == CURRENT->string */
      } /* while */
   } /* first if's else */
}

/*===============================================================*
 * Function: tree_search
 * Parameters: A character string constant.
 * Returns: True value if a character string has found from binary
 *          tree otherwise false, of course.
 *
 * Iterative binary tree search. while statement is repeated as 
 * long as there's nodes in the tree and searched character string 
 * is not found.
 *===============================================================*/
Boolean tree_search(char *str)
{
   P_TREENODE node;
   Boolean    found = FALSE;

   node = ROOT;
   while((node != NULL)&&(!found)){
      if(strcomp(str,node->string) < 0)
         node = node->left;
      else
      if(strcomp(str,node->string) > 0)
         node = node->right;
      else found = TRUE;
   } /* while */
   if(found) return(TRUE);
   else return(FALSE);
}

/*===============================================================*
 * Function: new_obj
 * Parameters: A character string constant.
 * Returns: Allocated new object for dynamic list.
 *
 * Memory allocation by using malloc(); function. Nothing serious. 
 * Only thing should be kept in mind is that object->string is 
 * also a pointer so it is sufficient reason to allocate memory 
 * also for it. Function gets a pointer to character string as a
 * parameter. After memory allocations the string will be copied 
 * to object->string and object->next is set as nil. Finally 
 * allocated object is returned.
 *===============================================================*/
P_LISTOBJ new_obj(char *str)
{
   char      *malloc();
   P_LISTOBJ object;

   object = (P_LISTOBJ) malloc(sizeof(LISTOBJ));
   if(object == NULL) syserror(5,error[0],"");
   object->string = malloc(strlen(str)+1);
   /* + 1 is needed to accommodate the terminating null character */
   if(object->string == NULL) syserror(5,error[0],"");
   strcpy(object->string,str);
   object->next = NULL;
   return(object);
}

/*===============================================================*
 * Function: add_list
 * Parameters: A character string constant.
 * Returns: Nothing.
 *
 * Purpose is to add a character string to non-ordered dynamic 
 * list. New object is created by using function new_obj. If first 
 * string in question, head and tail are same new object. In other 
 * case object will be added to tail.  Because P_LISTOBJ is a 
 * pointer to global structure, we don't have to return anything.
 *===============================================================*/
void add_list(char *str)
{
   P_LISTOBJ new_obj();

   if(HEAD == NULL)
      HEAD = TAIL = new_obj(str);       /* only member of list */
   else
      TAIL = TAIL->next = new_obj(str); /* add to tail */
}

/*===============================================================*
 * Funtion: new_buffer
 * Parameters: Nothing.
 * Returns: Nothing.
 *
 * Purpose is to create new buffer element. Memory allocation by 
 * using malloc(); function. buf->index set as 0 and buf->bext as 
 * nil. Finally allocated buffer object is returned. Because 
 * P_BUFFEROBJ is a pointer to global structure, we don't have to 
 * return anything.
 *===============================================================*/
P_BUFFEROBJ new_buffer()
{
   char        *malloc();
   P_BUFFEROBJ buf;

   buf = (P_BUFFEROBJ) malloc(sizeof(BUFFEROBJ));
   if(buf == NULL) syserror(5,error[0],"");
   buf->index = 0;
   buf->next = NULL;
   return(buf);
}

/*===============================================================*
 * Function: add_buffer
 * Parameters: A single character.
 * Returns: Nothing.
 *
 * Purpose is to add character into buffer. If buffer does not 
 * exist first and last buffer object is set to point to same new 
 * buffer created by new_buffer(); function. if buffer is full new 
 * buffer object is needed. 1) if already allocated buffer object
 * exist no need to allocate it again. character got as parameter 
 * is taken to buffer and next character in buffer is set as end 
 * of string marker. 2) must be bufsize-1 because of last->index+1 
 * 3). Because P_BUFFEROBJ is a pointer to global structure, we 
 * don't have to return anything.
 *===============================================================*/
void add_buffer(char character)
{
   if(FIRST == NULL)
      FIRST = LAST = new_buffer();
   if(LAST->index == BUFSIZE-1) /* 2 */
      if(LAST->next == NULL)
         LAST = LAST->next = new_buffer();
      else{ LAST = LAST->next; LAST->index = 0; } /* 1 */
   LAST->buffer[LAST->index] = character;
   LAST->buffer[LAST->index+1] = STREND; /* 3 */
   LAST->index++;
}

/*===============================================================*
 * Function: buffer_file
 * Parameters: A pointer to file and boolean value.
 * Returns: Nothing.
 * 
 * Purpose is to empty buffer to destination file Function gets 
 * file pointer and truth-value as parameters. If truth-value is 
 * true the whole buffer contents is copied to destination file.
 * Finally last is set to point to first and first's index is set 
 * to beginning.
 *===============================================================*/
void buffer_file(FILE *destfp, Boolean b)
{
   unsigned int i;
   P_BUFFEROBJ  buf;
   P_BUFFEROBJ  previous;


   buf = FIRST;
   while(buf != NULL){
      if(b){
         for(i=0;(i<BUFSIZE)&&(buf->buffer[i] != STREND);i++)
            fprintf(destfp,"%c",buf->buffer[i]);
      }
      previous = buf;
      buf = buf->next;
      free(previous);
   }
   LAST = FIRST = NULL;
}

/*===============================================================*
 * Function: gstrfstr
 * Parameters: A character string constant and two integer values.
 * Returns: A character string which is a partial string from 
 *          original got as a parameter.
 *
 * Purpose is to get string of unknown length from string. Function 
 * gets character string and two integer values as parameters. From 
 * value must be lower or equal compared with to value because 
 * purpose is to get partial character string from original 
 * character string. Function uses realloc(); function to 
 * reallocate memory when its needed.
 *===============================================================*/
char *gstrfstr(char *str, int from, int to)
{
   unsigned int i = 0;
   void   *realloc();
   size_t index;               /* size_t same as unsigned integer */
   size_t bufsize = 0;
   char   *buffer = NULL;

   for(index = from;index <= to;++index,++i){
      if(i+1>bufsize){
         if((buffer = realloc(buffer, bufsize+BLOCKSIZ+1)) == NULL)
            /* +1 because of end of string marker */
            syserror(5,error[0],"");
         bufsize += BLOCKSIZ;
      }
      buffer[i] = str[index];
   }
   buffer[i] = STREND;
   return(buffer);
}

/*===============================================================*
 * Function: brutalsearch
 * Parameters: Two character string constants.
 * Returns: Truth-value which is true if partial string1 is found 
 *          from string2, otherwise false.
 * 
 * Purpose is to search string1 from longer string2. function
 * returns naturally true truth-value if partial string1 is found 
 * from string2. Sure this is not very effective but it certainly
 * works properly. Algorithm divides string2 into parts size of 
 * string1 and compares partitial strings with string1. As soon as 
 * partitial string is same as string1 function returns true value 
 * else string1 is not partitial string of string2.
 *===============================================================*/
Boolean brutalsearch(char *str1, char *str2)
{
   char         *str;
   unsigned int len, from = 0, to;

   if(strlen(str2) >= strlen(str1)){
      len = strlen(str2)-1; /* first is 0 */
      to = strlen(str1)-1;
      while(to <= len){
         str = gstrfstr(str2,from,to);
         if(!strcomp(str1,str)) return(TRUE); /* found */
	 from++;to++;
      }
   }
   return(FALSE); /* not found */
}

/*===============================================================*
 * Function: list_search
 * Parameters: A character string constant.
 * Returns: Truth-value which is true if a character string is 
 *          found from dynamic list, otherwise false.
 * 
 * Purpose is to search character string from dynamic list. First 
 * temporary variable object is introduced. Object is set to point 
 * to head. as long as object not nil is object->string compared 
 * to character string got as function parameter.
 *===============================================================*/
Boolean list_search(char *str)
{
   P_LISTOBJ object; /* temporary variable */
   object = HEAD;

   while(object != NULL){
      if(brutalsearch(object->string,str)) return(TRUE); /* found */
	 object = object->next;
   }
   return(FALSE); /* not found */
}

/*===============================================================*
 * Function:gstrf
 * Parameters: A file pointer and a character.
 * Returns: A character string.
 * 
 * Purpose is to get string of unknown length from file. Function 
 * gets file pointer and end character as parameters. As long as 
 * file pointer does not point to the end character characters are 
 * taken to pointer type string character variable which is resized 
 * always when needed.
 *===============================================================*/
char *gstrf(FILE *srcfp, char ENDCHAR)
{
   char   c;
   void   *realloc();
   size_t index;
   size_t bufsize = 0;
   char   *buffer = NULL;

   for(index = 0;((c = getc(srcfp)) != ENDCHAR)&&(c != EOF);index++){
      if(index+1>bufsize){ /* if buffer is full, resize it */
         if((buffer = realloc(buffer, bufsize+BLOCKSIZ+1)) == NULL)
          /* +1 because of end of string marker */
            syserror(5,error[0],"");
         bufsize += BLOCKSIZ;
      }
      buffer[index] = c;
   }
   if(c==EOF)
      if((buffer=realloc(buffer, bufsize + 1)) == NULL)
         syserror(5,error[0],"");
   buffer[index] = STREND;
   return(buffer);
}

/*===============================================================*
 * Function: load_list
 * Parameters: A file pointer.
 * Returns: Nothing.
 *
 * Purpose is to add character strings from file to dynamic list.
 * As long as character string got from file is not end of string 
 * marker the string is added to dynamic list. Because in function
 * add_list() P_TREENODE is a pointer to global structure, we
 * don't have to return anything.
 *===============================================================*/
void load_list(FILE *srcfp)
{
   char *string;

   while(*(string = gstrf(srcfp,NEWLINE)) != STREND)
      add_list(string);
}

/*===============================================================*
 * Function: load_tree
 * Parameters: A file pointer.
 * Returns: Nothing.
 *
 * Purpose is to add character strings from file to binary tree.
 * As long as character string got from file is not end of string 
 * marker the string is added to binary tree. Because in function
 * add_tree P_TREENODE is a pointer to global structure, we
 * don't have to return anything.
 *===============================================================*/
void load_tree(FILE *srcfp)
{
   char *string;

   while(*(string = gstrf(srcfp,NEWLINE)) != STREND)
      add_tree(string);
}

/*===============================================================*
 * Function: add_tag_buffer
 * Parameters: Two characters, a character string and an integer 
 *             value.
 * Returns: Nothing.
 *
 * Purpose is to add string to buffer. First addition is an "open" 
 * character (in this case means '<' or '&' characters) then string 
 * is added and finally an "close" character (in this case means 
 * '>' or ';' characters). Because in function add_list() 
 * P_BUFFEROBJ is a pointer to global structure, we don't have to 
 * return anything.
 *===============================================================*/
void add_tag_buffer(char open, char close, char *string, int length)
{
   unsigned int i;
   add_buffer(open);
   for(i = 0;i<length;i++) add_buffer(string[i]);
   add_buffer(close);
}

/*===============================================================*
 * Function: load_entities
 * Parameters: Nothing.
 * Returns: Nothing.
 * 
 * Purpose is to take inner entities to binary tree. Very simple 
 * repetion statement. entity[i] points to pointer array *entity[] 
 * (introduced above as a global variable). Because in function 
 * add_tree() P_TREENODE is a pointer to global structure, we
 * don't have to return anything.
 *===============================================================*/
void load_entities()
{
   unsigned int i=0;

   while(strcomp(entity[i],"the_end")){
      add_tree(entity[i]);
      i++;
   }
}

/*===============================================================*
 * Function: freetree
 * Parameters: Pointer to tree node.
 * Returns: Nothing. 
 *
 * Recursive routine which passes through binary tree and makes
 * blocks previously allocated available for further allocation.
 *===============================================================*/
P_TREENODE freetree(P_TREENODE node)
{
   if(node != NULL){
      freetree(node->left);
      free(node->string);
      free(node);
      freetree(node->right);
   }
}

/*===============================================================*
 * Function: freelist
 * Parameters: Pointer to list object.
 * Returns: Nothing.
 * 
 * Passes through dynamic list and makes blocks previously 
 * allocated available for further allocation.
 *===============================================================*/
void freelist(P_LISTOBJ obj)
{
   P_LISTOBJ previous; 

   while(obj != NULL){
      previous = obj;
      obj = obj->next;
      free(previous->string);
      free(previous);
   }
}
